/**
 * Module dependencies.
 */
var passport = require('passport')
  , https = require('https')
  , querystring = require('querystring')
  , util = require('util')


/**
 * `Strategy` constructor.
 *
 * The BrowserID authentication strategy authenticates requests using the
 * BrowserID JavaScript API and Verified Email Protocol (VEP).
 *
 * BrowserID provides a federated and decentralized universal login system for
 * the web, based on email addresses as an identity token.  Authenticating in
 * this this manner involves a sequence of events, including prompting the user,
 * via their user agent, for an assertion of email address ownership.  Once this
 * assertion is obtained, it can be verified and the user can be authenticated.
 *
 * Applications must supply a `validate` callback which accepts an `email`
 * address, and then calls the `done` callback supplying a `user`, which should
 * be set to `false` if the credentials are not valid.  If an exception occured,
 * `err` should be set.
 *
 * Options:
 *   - `audience`        the website requesting and verifying an identity assertion
 *   - `assertionField`  field name where the assertion is found, defaults to 'assertion'
 *
 * Examples:
 *
 *     passport.use(new BrowserIDStrategy({
 *         audience: 'http://www.example.com'
 *       },
 *       function(email, done) {
 *         User.findByEmail(email, function (err, user) {
 *           done(err, user);
 *         });
 *       }
 *     ));
 *
 * @param {Object} options
 * @param {Function} validate
 * @api public
 */
function Strategy(options, validate) {
  if (!options.audience) throw new Error('BrowserID authentication requires an audience option');
  if (!validate) throw new Error('BrowserID authentication strategy requires a validate function');
  
  passport.Strategy.call(this);
  this.name = 'browserid';
  this._validate = validate;
  
  this._audience = options.audience;
  this._assertionField = options.assertionField || 'assertion';
  
  // options used to inject mock objects for testing purposes
  this._https = options.transport || https;
}

/**
 * Inherit from `passport.Strategy`.
 */
util.inherits(Strategy, passport.Strategy);


/**
 * Authenticate request by using browserid.org as a trusted secondary authority
 * for verifying email assertions.
 *
 * @param {Object} req
 * @api protected
 */
Strategy.prototype.authenticate = function(req) {
  var self = this;
  
  if (!req.body || !req.body[this._assertionField]) {
    return this.fail();
  }

  var assertion = req.body[this._assertionField];
  
  var query = querystring.stringify({ assertion: assertion, audience: this._audience });
  var headers = {};
  headers['Host'] = 'browserid.org';
  headers['Content-Type'] = 'application/x-www-form-urlencoded';
  headers['Content-Length'] = query.length;
  
  var options = {
    host: 'browserid.org',
    port: 443,
    path: '/verify',
    method: 'POST',
    headers: headers
  };
  var req = this._https.request(options, function(res) {
    var data = '';
    res.on('data', function(chunk) {
      data += chunk;
    });
    res.on('end', function() {
      try {
        var result = JSON.parse(data);
        if (result.status === 'okay') {
          return verified(result)
        } else {
          // TODO: Implement a mechanism in Passport to provide errors when
          //       authentication failures occur.
          // return self.fail(new Error(result.reason));
          return self.fail();
        }
      } catch(err) {
        return self.error(err);
      }
    });
    res.on('error', function(err) {
      return self.error(err);
    });
  });
  req.end(query, 'utf8');
  
  // TODO: Check that the audience matches this server, according to the Host
  //       header.  Also, implement an option to disable this check (defaulting
  //       to false).
  
  function verified(result) {
    self._validate(result.email, function(err, user) {
      if (err) { return self.error(err); }
      if (!user) { return self.fail(); }
      self.success(user);
    });
  }
}


/**
 * Expose `Strategy`.
 */ 
module.exports = Strategy;
